import re
from dfsql.engine import pd
from collections import Iterable

from dfsql.utils import (is_modin, is_numeric, is_booly, is_stringy, raise_bad_inputs, raise_bad_outputs,
                               TwoArgsMixin, OneArgMixin, StringInputMixin, NumericInputMixin, BoolInputMixin,
                               BoolOutputMixin, StringOutputMixin, NumericOutputMixin)


class BaseFunction:
    name = None

    # Fixes an issue with Modin internals trying to get the __name__ of aggregation functions
    @property
    def __name__(self):
        return self.name

    def assert_args(self, args):
        super().assert_args(args)

    def assert_output(self, out):
        super().assert_output(out)

    def get_output(self, args):
        return None

    def __call__(self, *args):
        self.assert_args(args)
        output = self.get_output(args)
        self.assert_output(output)
        return output

# Explanation on how this function definition works:
# https://stackoverflow.com/a/40187463/1571481

# Boolean functions


class And(BaseFunction, TwoArgsMixin, BoolInputMixin, BoolOutputMixin):
    name = 'and'

    def get_output(self, args):
        if is_modin(args[0]) and is_modin(args[1]):
            return (args[0] * args[1]).astype(bool)
        return args[0] and args[1]


class Or(BaseFunction, TwoArgsMixin, BoolInputMixin, BoolOutputMixin):
    name = 'or'

    def get_output(self, args):
        if is_modin(args[0]) and is_modin(args[1]):
            return (args[0] + args[1]).astype(bool)
        return args[0] or args[1]


class Not(BaseFunction, BoolOutputMixin):
    name = 'not'

    def assert_args(self, args):
        if len(args) != 1:
            raise_bad_inputs(self)

        if not (is_modin(args[0])
                or isinstance(args[0], bool)
                or (args[0] in (0, 1))):
            raise_bad_inputs(self)

    def get_output(self, args):
        if is_modin(args[0]):
            return ~args[0]
        return not args[0]


class Is(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = 'is'

    def get_output(self, args):
        if is_modin(args[0]) or is_modin(args[1]):
            if args[0] is None or args[1] is None:
                # IS NULL
                target = args[0]
                if args[0] is None:
                    target = args[1]

                return pd.isnull(target)
            elif args[0] is True or args[0] is False or args[1] is True or args[1] is False and is_modin(args[0]):
                # IS [TRUE|FALSE]
                return args[0] == args[1]

        return args[0] is args[1]


class IsNot(Is):
    name = 'is not'

    def get_output(self, args):
        out = super().get_output(args)
        if is_modin(out):
            return ~out
        else:
            return not out


class Equals(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '='

    def get_output(self, args):
        return args[0] == args[1]


class NotEquals(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '!='

    def get_output(self, args):
        return args[0] != args[1]


class Greater(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '>'

    def get_output(self, args):
        return args[0] > args[1]


class GreaterEqual(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '>='

    def get_output(self, args):
        return args[0] >= args[1]


class Less(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '<'

    def get_output(self, args):
        return args[0] < args[1]


class LessEqual(BaseFunction, TwoArgsMixin, BoolOutputMixin):
    name = '<='

    def get_output(self, args):
        return args[0] <= args[1]


class In(BaseFunction, BoolOutputMixin):
    name = 'in'

    def assert_args(self, args):
        if not isinstance(args[1], Iterable):
            raise_bad_inputs(self)

    def get_output(self, args):
        if is_modin(args[0]):
            return args[0].isin(args[1].values)
        return args[0] in args[1]


# class IsNull(BaseFunction, OneArgMixin, BoolOutputMixin):
#     name = 'is null'
#
#     def get_output(self, args):
#         return pd.isnull(args[0])
#
#
# class IsNotNull(BaseFunction, OneArgMixin, BoolOutputMixin):
#     name = 'is not null'
#
#     def get_output(self, args):
#         return ~pd.isnull(args[0])
#
#
# class IsTrue(BaseFunction, OneArgMixin, BoolOutputMixin):
#     name = 'is true'
#
#     def get_output(self, args):
#         if is_modin(args[0]):
#             return args[0] == True
#         return args[0] is True
#
#
# class IsFalse(BaseFunction, OneArgMixin, BoolOutputMixin):
#     name = 'is false'
#
#     def get_output(self, args):
#         if is_modin(args[0]):
#             return args[0] == False
#         return args[0] is False

# Arithmetic functions


class Plus(BaseFunction, TwoArgsMixin, NumericInputMixin, NumericOutputMixin):
    name = '+'

    def get_output(self, args):
        return pd.to_numeric(args[0] + args[1])


class Minus(BaseFunction, NumericOutputMixin):
    name = '-'

    def assert_args(self, args):
        if not (len(args) == 1 or len(args) == 2):
            raise_bad_inputs(self)

        if len(args) == 2:
            if not ((is_modin(args[0]) and is_modin(args[1]))
                    or (is_numeric(args[0]) and is_numeric(args[1]))):
                raise_bad_inputs(self)

        if len(args) == 1:
            if not (is_modin(args[0]) or (is_numeric(args[0]))):
                raise_bad_inputs(self)

    def get_output(self, args):
        if len(args) == 1:
            return pd.to_numeric(-args[0])
        return pd.to_numeric(args[0] - args[1])


class Multiply(BaseFunction, TwoArgsMixin, NumericInputMixin, NumericOutputMixin):
    name = '*'

    def get_output(self, args):
        return pd.to_numeric(args[0] * args[1])


class Divide(BaseFunction, TwoArgsMixin, NumericInputMixin, NumericOutputMixin):
    name = '/'

    def get_output(self, args):
        return pd.to_numeric(args[0] / args[1])


class Modulo(BaseFunction, TwoArgsMixin, NumericInputMixin, NumericOutputMixin):
    name = '%'

    def get_output(self, args):
        return pd.to_numeric(args[0] % args[1])


class Power(BaseFunction, TwoArgsMixin, NumericInputMixin, NumericOutputMixin):
    name = '^'

    def get_output(self, args):
        return pd.to_numeric(args[0] ** args[1])

# String functions


class StringConcat(BaseFunction, TwoArgsMixin, StringInputMixin, StringOutputMixin):
    name = "||"

    def get_output(self, args):
        return args[0] + args[1]


class StringLower(BaseFunction, OneArgMixin, StringInputMixin, StringOutputMixin):
    name = "lower"

    def get_output(self, args):
        if isinstance(args[0], str):
            return args[0].lower()
        return args[0].apply(lambda x: x.lower())


class StringUpper(BaseFunction, OneArgMixin, StringInputMixin, StringOutputMixin):
    name = "upper"

    def get_output(self, args):
        if isinstance(args[0], str):
            return args[0].upper()
        return args[0].apply(lambda x: x.upper())


class Like(BaseFunction, TwoArgsMixin, StringInputMixin, BoolOutputMixin):
    name = "like"

    def get_output(self, args):
        def matcher(inp, pattern):
            match = re.match(pattern, inp)
            return True if match else False

        if is_modin(args[0]):
            return args[0].apply(matcher, args=(args[1],))
        return matcher(args[0], args[1])

# Aggregate functions


class AggregateFunction(BaseFunction, OneArgMixin):
    string_repr = None # for pandas group by

    def assert_output(self, output):
        pass

    @classmethod
    def string_or_callable(cls):
        if cls.string_repr:
            return cls.string_repr
        return cls()


class Mean(AggregateFunction):
    name = 'avg'
    string_repr = 'mean'


class Sum(AggregateFunction):
    name = 'sum'
    string_repr = 'sum'


class Count(AggregateFunction):
    name = 'count'
    string_repr = 'count'


class CountDistinct(AggregateFunction):
    name = 'count_distinct'
    string_repr = 'nunique'


class Max(AggregateFunction):
    name = 'max'
    string_repr = 'max'


class Min(AggregateFunction):
    name = 'min'
    string_repr = 'min'

OPERATIONS = (
    And, Or, Not,

    Equals, NotEquals, Greater, GreaterEqual, Less, LessEqual, Is, IsNot,

    Plus, Minus, Multiply, Divide, Modulo, Power,

    StringConcat, StringLower, StringUpper, Like,

    In,

    # IsNull, IsNotNull, IsTrue, IsFalse
)

OPERATION_MAPPING = {
    op.name: op for op in OPERATIONS
}
OPERATION_MAPPING['<>'] = NotEquals

AGGREGATE_FUNCTIONS = (
    Sum, Mean, Count, CountDistinct, Max, Min,
)

AGGREGATE_MAPPING = {
    op.name: op for op in AGGREGATE_FUNCTIONS
}


def is_supported(op_name):
    return op_name.lower() in OPERATION_MAPPING or op_name.lower() in AGGREGATE_MAPPING



from functools import lru_cache


class BaseCache:
    def get(self, table):
        pass

    def clear(self):
        pass


class DoNothingCache(BaseCache):
    pass


class MemoryCache(BaseCache):
    def __init__(self, maxsize=None):
        decorated_get = lru_cache(maxsize=maxsize)(self.get)
        setattr(self, 'get', decorated_get)

    def clear(self):
        self.get.cache_clear()

    def get(self, table):
        df = table.fetch_and_preprocess()
        return df



import logging
import re
from dfsql import sql_query
from dfsql.engine import pd as pd_engine
import warnings
import pandas
from pandas.core.accessor import CachedAccessor


@pandas.api.extensions.register_dataframe_accessor("sql")
class SQLAccessor:
    def __init__(self, pandas_obj):
        self._obj = pd_engine.DataFrame(pandas_obj)

    def maybe_add_from_to_query(self, sql_query, table_name):
        """Inserts "FROM temp" into every SELECT clause in query that does not have a FROM clause."""
        sql_query = sql_query.lower()
        sql_query = sql_query.replace("(", " ( ").replace(")", " ) ")
        insert_positions = []
        for m in re.finditer('select', sql_query):
            select_pos = m.start()

            str_after_select = sql_query[select_pos:]
            words_after_select = str_after_select.split(' ')

            keywords = ['where', 'group', 'having', 'order', 'limit', 'offset']
            need_to_insert_from = True
            insert_pos = len(str_after_select)

            parentheses_count = 0
            for word in words_after_select:
                if word == 'from':
                    need_to_insert_from = False
                    break

                if word == '(':
                    parentheses_count += 1
                elif word == ')':
                    if parentheses_count == 0:
                        insert_pos = str_after_select.find(word)
                        break
                    else:
                        parentheses_count -= 1
                if word in keywords:
                    insert_pos = str_after_select.find(word)
                    break
            if not need_to_insert_from:
                continue
            insert_pos = select_pos + insert_pos

            insert_positions.append(insert_pos)
        insert_text = f' from {table_name} '
        new_query = ''
        last_pos = None
        for pos in insert_positions:
            new_query += sql_query[last_pos:pos] + insert_text
            last_pos = pos

        new_query += sql_query[last_pos:]
        return new_query

    def __call__(self, sql, *args, **kwargs):
        table_name = 'temp'
        sql = self.maybe_add_from_to_query(sql, table_name=table_name)
        kwargs.update({table_name: self._obj})
        return sql_query(sql, *args, **kwargs)

try:
    import modin.pandas as mpd

    def register_modin_accessor(name, cls):
        def decorator(accessor):
            if hasattr(cls, name):
                warnings.warn(
                    f"registration of accessor {repr(accessor)} under name "
                    f"{repr(name)} for type {repr(cls)} is overriding a preexisting "
                    f"attribute with the same name.",
                    UserWarning,
                    stacklevel=2,
                )

            setattr(cls, name, CachedAccessor(name, accessor))
            return accessor

        return decorator


    def register_modin_dataframe_accessor(name):
        from modin.pandas import DataFrame
        return register_modin_accessor(name, DataFrame)

    register_modin_dataframe_accessor("sql")(SQLAccessor)
except ImportError:
    warnings.warn('Modin not found, dfsql Modin dataframe extensions not loaded', UserWarning)



class DfsqlException(Exception):
    pass


class SQLParsingException(DfsqlException):
    pass


class CommandException(DfsqlException):
    pass


class QueryExecutionException(DfsqlException):
    pass




from dfsql.engine import pd
from dfsql.exceptions import QueryExecutionException


def raise_bad_inputs(func):
    raise QueryExecutionException(f'Invalid inputs for function {func.name}')


def raise_bad_outputs(func):
    raise QueryExecutionException(f'Invalid outputs produced by function {func.name}')


def is_modin(thing):
    return (isinstance(thing, pd.Series) or isinstance(thing, pd.DataFrame))


def is_booly(thing):
    if ((is_modin(thing))
            or isinstance(thing, bool)
            or (int(thing) in (0, 1))):
        return True
    return False


def is_numeric(thing):
    if ((is_modin(thing) and thing.dtype.name != 'object')
            or isinstance(thing, int) or isinstance(thing, float)):
        return True
    return False


def is_stringy(thing):
    if ((is_modin(thing) and thing.dtype.name in ('string', 'object'))
            or isinstance(thing, str)):
        return True
    return False


class TwoArgsMixin:
    def assert_args(self, args):
        if len(args) != 2:
            raise_bad_inputs(self)


class OneArgMixin:
    def assert_args(self, args):
        if len(args) != 1:
            raise_bad_inputs(self)


class BoolInputMixin:
    def assert_args(self, args):
        if not all([is_booly(arg) for arg in args]):
            raise_bad_inputs(self)


class BoolOutputMixin:
    def assert_output(self, output):
        if not is_booly(output):
            raise_bad_outputs(self)


class NumericInputMixin:
    def assert_args(self, args):
        if not all([is_numeric(arg) for arg in args]):
            raise_bad_inputs(self)


class NumericOutputMixin:
    def assert_output(self, output):
        if not is_numeric(output):
            raise_bad_outputs(self)


class StringInputMixin:
    def assert_args(self, args):
        if not all([is_stringy(arg) for arg in args]):
            raise_bad_inputs(self)


class StringOutputMixin:
    def assert_output(self, output):
        if not is_stringy(output):
            raise_bad_outputs(self)




import re
from dfsql.engine import pd
import numpy as np
import os


def preprocess_column_name(text):
    text = re.sub(' +', ' ', text)
    text = "".join(c for c in text if c.isalnum() or c in (' ', '_', '-')).strip().replace('.', '_').replace(' ', '_').replace('-', '_').lower()
    return text


def make_preprocessing_dict(df):
    rename = {col: preprocess_column_name(col) for col in df.columns}
    empty_rows = df.index[df.isnull().all(axis=1)].values.tolist()
    drop_columns = df.columns[df.isnull().all(axis=0)].values.tolist()
    return dict(rename=rename, empty_rows=empty_rows, drop_columns=drop_columns)


def preprocess_dataframe(df, rename, empty_rows, drop_columns):
    df.columns = [rename[col] for col in df.columns]
    df = df.drop(drop_columns, axis=1)
    df = df.drop(empty_rows, axis=0)
    df.index = range(len(df))
    df = df.convert_dtypes()
    return df


class Table:
    def __init__(self, name, *args, preprocessing_dict=None, cache=None, **kwargs):
        self.name = name
        self.preprocessing_dict = preprocessing_dict
        self.cache = cache

    def __hash__(self):
        return hash(self.name)

    def fetch_dataframe(self):
        pass

    def preprocess_dataframe(self, df):
        return preprocess_dataframe(df, **self.preprocessing_dict)

    def fetch_and_preprocess(self):
        df = self.fetch_dataframe()
        if self.preprocessing_dict:
            df = self.preprocess_dataframe(df)
        return df

    @property
    def dataframe(self):
        if self.cache:
            return self.cache.get(self)

        return self.fetch_and_preprocess()

    def to_json(self):
        return dict(
            type=self.__class__.__name__,
            name=self.name,
            preprocessing_dict=self.preprocessing_dict,
        )

    @staticmethod
    def from_json(json):
        cls = {
            'Table': Table,
            'FileTable': FileTable
        }[json['type']]
        return cls(**json)


class FileTable(Table):
    def __init__(self, *args, fpath, **kwargs):
        super().__init__(*args, **kwargs)
        self.fpath = fpath

    def fetch_dataframe(self):
        return pd.read_csv(self.fpath)

    @classmethod
    def from_file(cls, path, clean=True):
        fpath = os.path.join(path)
        fname = '.'.join(os.path.basename(fpath).split('.')[:-1])

        table = cls(name=fname, fpath=fpath)
        df = table.fetch_dataframe()
        if clean:
            preprocessing_dict = make_preprocessing_dict(df)
            table.preprocessing_dict = preprocessing_dict

        return table

    def to_json(self):
        json = super().to_json()
        json['fpath'] = self.fpath
        return json



from confi import BaseEnvironConfig, ConfigField, ConfigError, BooleanConfig
from distutils.util import strtobool
import logging


def true_if_modin_installed():
    try:
        import modin
        logging.info(
            "Detected Modin and an explicit USE_MODIN value was not provided. Modin will be used for dfsql operations.")
        return True
    except ImportError:
        return False


class Configuration(BaseEnvironConfig):
    USE_MODIN = BooleanConfig(default=true_if_modin_installed)



import os
from dfsql.engine import pd
import json

from dfsql.cache import  MemoryCache
from dfsql.exceptions import QueryExecutionException
from dfsql.functions import OPERATION_MAPPING, AGGREGATE_MAPPING
from dfsql.commands import try_parse_command
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import (Select, Identifier, Constant, Operation, Function, Join, BinaryOperation, TypeCast,
                                    Tuple, NullConstant, Star)
from dfsql.table import Table, FileTable, preprocess_column_name


def get_modin_operation(sql_op):
    op = OPERATION_MAPPING.get(sql_op.lower())
    if not op:
        raise(QueryExecutionException(f'Unsupported operation: {sql_op}'))
    return op()


def get_aggregation_operation(sql_op):
    op = AGGREGATE_MAPPING.get(sql_op.lower())
    if not op:
        raise(QueryExecutionException(f'Unsupported operation: {sql_op}'))
    return op.string_or_callable()


def cast_type(obj, type_name):
    if not hasattr(obj, 'astype'):
        obj = pd.Series(obj)
    return obj.astype(type_name)


class DataSource:
    def __init__(self, metadata_dir, tables=None, cache=None, custom_functions=None):
        self.metadata_dir = metadata_dir

        if not os.path.exists(self.metadata_dir):
            os.makedirs(self.metadata_dir, exist_ok=True)

        tables = {t.name.lower(): t for t in tables} if tables else {}
        self.tables = None

        self.load_metadata()
        if self.tables and tables:
            raise QueryExecutionException(f'Table metadata already exists in directory {metadata_dir}, but tables also passed to DataSource constructor. '
                            f'\nEither load the previous metadata by omitting the tables argument, or explicitly overwrite old metadata by using DataSource.create_new(metadata_dir, tables).')
        elif not self.tables:
            self.tables = tables

        self.save_metadata()

        self.set_cache(cache or MemoryCache())

        self._query_scope = {}
        
        self.custom_functions = custom_functions or {}

    def set_cache(self, cache):
        self.cache = cache
        for tname, table in self.tables.items():
            table.cache = self.cache

    @property
    def query_scope(self):
        """Stores aliases and tables available to a select during execution"""
        return self._query_scope

    def clear_query_scope(self):
        self._query_scope = {}

    @classmethod
    def create_new(cls, metadata_dir, tables=None):
        cls.clear_metadata(metadata_dir)
        return cls(metadata_dir, tables=tables)

    @classmethod
    def clear_metadata(cls, metadata_dir):
        if os.path.exists(os.path.join(metadata_dir, 'datasource_tables.json')):
            os.remove(os.path.join(metadata_dir, 'datasource_tables.json'))

    def add_table_from_file(self, path, clean=True):
        table = FileTable.from_file(path, clean=clean)
        self.add_table(table)

    @staticmethod
    def from_dir(metadata_dir, files_dir_path):
        files = os.listdir(files_dir_path)
        ds = DataSource(metadata_dir=metadata_dir)
        for f in files:
            if f.endswith('.csv'):
                fpath = os.path.join(files_dir_path, f)
                ds.add_table_from_file(fpath)

        if not ds.tables:
            raise(QueryExecutionException(f'Directory {files_dir_path} does not contain any spreadsheet files'))
        return ds

    def load_metadata(self):
        if not os.path.exists(os.path.join(self.metadata_dir, 'datasource_tables.json')):
            return

        new_tables = {}
        with open(os.path.join(self.metadata_dir, 'datasource_tables.json'), 'r') as f:
            table_data = json.load(f)

        for tname, table_json in table_data.items():
            new_tables[tname] = Table.from_json(table_json)

        self.tables = new_tables

    def save_metadata(self, overwrite=True):
        if not os.path.exists(self.metadata_dir):
            os.makedirs(self.metadata_dir)

        if not os.access(self.metadata_dir, os.W_OK):
            raise QueryExecutionException(f'Directory {self.metadata_dir} not writable')

        tables_dump = {
            tname: table.to_json() for tname, table in self.tables.items()
        }

        if not overwrite and os.path.exists(os.path.join(self.metadata_dir, 'datasource_tables.json')):
            raise QueryExecutionException('Table metadata already exists, but overwrite is False.')

        with open(os.path.join(self.metadata_dir, 'datasource_tables.json'), 'w') as f:
            f.write(json.dumps(tables_dump))

    def __contains__(self, table_name):
        return table_name in self.tables

    def register_function(self, name, func):
        self.custom_functions[name] = func

    def add_table(self, table):
        if self.tables.get(table.name):
            raise QueryExecutionException(f'Table {table.name} already exists in data source, use DROP TABLE to remove it if you want to recreate it.')
        self.tables[table.name] = table
        self.save_metadata()

    def drop_table(self, name):
        del self.tables[name]
        self.save_metadata()

    def execute_command(self, command):
        return command.execute(self)

    def query(self, sql):
        command = try_parse_command(sql)
        if command:
            return self.execute_command(command)
        query = parse_sql(sql)
        return self.execute_query(query, reduce_output=True)

    def execute_table_identifier(self, query):
        table_name = query.parts_to_str()
        if table_name not in self:
            raise QueryExecutionException(f'Unknown table {table_name}')
        else:
            df = self.tables[table_name].dataframe
            self.query_scope[table_name] = df
            return df

    def execute_constant(self, query):
        if isinstance(query, NullConstant):
            return None
        value = query.value
        return value

    def execute_operation(self, query, df):
        args = [self.execute_select_target(arg, df) for arg in query.args]
        op_func = self.custom_functions.get(query.op.lower())
        if not op_func:
            op_func = get_modin_operation(query.op.lower())
        result = op_func(*args)
        return result

    def execute_column_identifier(self, query, df):
        name_components = query.parts

        if len(name_components) == 1:
            full_column_name = preprocess_column_name(name_components[0])
            if full_column_name in df.columns:
                return df[full_column_name]
        elif len(name_components) == 2:
            table_name, column_name = name_components
            column_name = preprocess_column_name(column_name)

            # If it's a join or a subquery
            join_column_name = f'{table_name}.{column_name}'
            if join_column_name in df:
                return df[join_column_name]

            if table_name and not table_name in self.query_scope:
                raise QueryExecutionException(f"Table name {table_name} not in scope.")
            table = self.query_scope[table_name]

            if column_name in table.columns:
                column = table[column_name]
                column.name = query.parts_to_str()
                return column
        else:
            raise QueryExecutionException(f"Too many name components: {query.parts}")
        raise QueryExecutionException(f"Column {query.parts_to_str()} not found.")

    def execute_type_cast(self, query, df):
        type_name = query.type_name
        arg = self.execute_select_target(query.arg, df)
        return cast_type(arg, type_name)

    def execute_select_target(self, query, df):
        if isinstance(query, Identifier):
            return self.execute_column_identifier(query, df)
        elif isinstance(query, Operation):
            return self.execute_operation(query, df)
        elif isinstance(query, TypeCast):
            return self.execute_type_cast(query, df)

        return self.execute_query(query, reduce_output=True)

    def resolve_select_target_col_name(self, target):
        col_name = target.alias
        if not col_name:
            if isinstance(target, Identifier):
                col_name = target.parts_to_str()
            else:
                col_name = str(target)
        return col_name

    def execute_select_targets(self, targets, source_df):
        out_names = []

        iterable_names = []
        iterable_columns = []

        scalar_names = []
        scalar_values = []
        # Expand star
        for i, target in enumerate(targets):
            if isinstance(target, Star):
                targets = targets[:i] + [Identifier(colname) for colname in
                                                     source_df.columns] + targets[i + 1:]
                break

        for target in targets:
            col_name = self.resolve_select_target_col_name(target)
            out_names.append(col_name)

            select_target_result = self.execute_select_target(target, source_df)
            if isinstance(select_target_result, pd.Series):
                iterable_names.append(col_name)
                iterable_columns.append(select_target_result)
            else:
                scalar_names.append(col_name)
                scalar_values.append(select_target_result)

        # Add columns first, then scalars, so the dataframe has proper index in the end
        out_columns = {}
        for i, col_name in enumerate(iterable_names):
            out_columns[col_name] = list(iterable_columns[i])
        out_df = pd.DataFrame.from_dict(out_columns)
        for i, col_name in enumerate(scalar_names):
            if out_df.empty:
                out_df[col_name] = [scalar_values[i]]
            else:
                out_df[col_name] = scalar_values[i]
        out_df = out_df[out_names]
        return out_df

    def execute_select_groupby_targets(self, targets, source_df, group_by):
        final_out_column_names = []
        out_column_names = []
        out_columns = []

        agg = {}

        group_by_cols = [q.parts_to_str() for q in group_by if q != Constant(True)]
        for target in targets:
            col_df_name = target.alias
            col_name = self.resolve_select_target_col_name(target)

            if col_name in agg:
                raise QueryExecutionException(f'Duplicate column name {col_name}. Provide an alias to resolve ambiguity.')

            if isinstance(target, Identifier):
                col_df_name = target.parts_to_str()

            if isinstance(target, Function):
                arg = target.args[0]
                if not isinstance(arg, Identifier):
                    raise QueryExecutionException(f'The argument of an aggregate function must be a column, found: {str(arg)}')
                arg = arg.parts_to_str()

                func_name = target.op.lower()
                if target.distinct:
                    func_name = f'{target.op.lower()}_distinct'
                modin_op = self.custom_functions.get(func_name)
                if not modin_op:
                    modin_op = get_aggregation_operation(func_name)

                agg[col_name] = (arg, modin_op)
            else:
                if col_name not in group_by_cols and col_df_name not in group_by_cols:
                    raise QueryExecutionException(f'Column {col_df_name}({col_name}) not found in GROUP BY clause')

            final_out_column_names.append(col_name)

        if isinstance(source_df, pd.Series):
            source_df = pd.DataFrame(source_df)

        if isinstance(source_df, pd.DataFrame):
            temp_df = pd.DataFrame(source_df)
            temp_df['__dummy__'] = 0
            source_df = temp_df.groupby('__dummy__')
        aggregate_result = source_df.agg(**agg)

        for col_index in aggregate_result.reset_index().columns:
            if col_index not in final_out_column_names:
                continue
            out_column_names.append(col_index)
            out_columns.append(aggregate_result.reset_index()[col_index].values)

        out_dict = {col: values for col, values in zip(out_column_names, out_columns)}
        out_df = pd.DataFrame.from_dict(out_dict)
        return out_df

    def execute_order_by(self, order_by, df):
        fields = [s.field.parts_to_str() for s in order_by]
        sort_orders = [s.direction != 'DESC' for s in order_by]
        df = df.sort_values(by=fields, ascending=sort_orders)
        return df

    def execute_select(self, query, reduce_output=False):
        from_table = []
        if query.from_table:
            from_table = self.execute_from_query(query.from_table)

        source_df = from_table

        if query.where:
            index = self.execute_operation(query.where, source_df)
            source_df = source_df[index.values]
        group_by = False

        if query.group_by is None:
            # Check for implicit group by
            non_agg_functions = []
            agg_functions = []
            for target in query.targets:
                if isinstance(target, Function) and target.op.lower() in AGGREGATE_MAPPING:
                    agg_functions.append(target)
                else:
                    non_agg_functions.append(target)

            if not non_agg_functions and agg_functions:
                query.group_by = [Constant(True)]
            elif non_agg_functions and agg_functions:
                raise(QueryExecutionException(f'Can\'t process a mix of aggregation functions and non-aggregation functions with no GROUP BY clause.'))
        if query.group_by:
            group_by = True
            source_df = self.execute_groupby_queries(query.group_by, source_df)

        if group_by == False:
            out_df = self.execute_select_targets(query.targets, source_df)
        else:
            out_df = self.execute_select_groupby_targets(query.targets, source_df, query.group_by)

        if query.having:
            if group_by == False:
                raise QueryExecutionException('Can\'t execute HAVING clause with no GROUP BY clause.')
            index = self.execute_operation(query.having, out_df)
            out_df = out_df[index]

        if query.distinct:
            out_df = out_df.drop_duplicates()

        if query.offset:
            offset = self.execute_query(query.offset)
            out_df = out_df.iloc[offset:, :]

        if query.order_by:
            out_df = self.execute_order_by(query.order_by, out_df)

        if query.limit:
            limit = self.execute_query(query.limit)
            out_df = out_df.iloc[:limit, :]

        self.clear_query_scope()

        # Turn tables into Series or constants if needed, for final returning
        if reduce_output:
            if out_df.shape == (1, 1): # Just one value returned
                return out_df.values[0][0]
            elif out_df.shape[1] == 1:
                return out_df[out_df.columns[0]]
        return out_df

    def execute_join(self, query):
        join_type = query.join_type
        join_type = {'INNER JOIN': 'inner', 'LEFT JOIN': 'left', 'RIGHT JOIN': 'right', 'FULL JOIN': 'outer'}[join_type]

        left = query.left
        if isinstance(left, Identifier):
            left = self.execute_table_identifier(left)
        else:
            left = self.execute_query(left)

        right = query.right
        if isinstance(right, Identifier):
            right = self.execute_table_identifier(right)
        else:
            right = self.execute_query(right)

        condition = query.condition
        if isinstance(condition, BinaryOperation):
            left_on = condition.args[0]
            right_on = condition.args[1]
        else:
            raise QueryExecutionException(f'Invalid join condition {condition.op}')
        left_name = query.left.alias if query.left.alias else query.left.parts_to_str()
        right_name = query.right.alias if query.right.alias else query.right.parts_to_str()
        left_on, right_on = left_on if left_on.parts[0] in left_name else right_on, \
                            right_on if right_on.parts[0] in right_name else left_on

        left_on = left_on.parts[-1]
        right_on = right_on.parts[-1]
        out_df = pd.merge(left, right, how=join_type, left_on=[left_on], right_on=[right_on], suffixes=('_x', '_y'))
        renaming = {f'{left_on}_x': left_on, f'{right_on}_y': right_on}

        for col in out_df.columns:
            if col in renaming:
                continue

            if '_x' in col:
                pure_col_name = col.replace('_x', '')
                renaming[col] = f'{left_name}.{pure_col_name}'
            elif '_y' in col:
                pure_col_name = col.replace('_y', '')
                renaming[col] = f'{right_name}.{pure_col_name}'

        out_df = out_df.rename(renaming, axis=1)
        return out_df

    def execute_from_query(self, query):
        if isinstance(query, Identifier):
            df = self.execute_table_identifier(query)
        elif isinstance(query, Join):
            df = self.execute_join(query)
        else:
            df = self.execute_query(query)

        if query.alias:
            self.query_scope[query.alias] = df

        return df

    def execute_groupby_queries(self, queries, df):
        col_names = []

        if len(queries) == 1 and queries[0] == Constant(True):
            return df

        for query in queries:
            if isinstance(query, Identifier):
                column = self.execute_column_identifier(query, df)
                col_names.append(column.name)
            else:
                raise QueryExecutionException(f"Don't know how to aggregate by {str(query)}")
        return df.groupby(col_names)

    def execute_query(self, query, reduce_output=False):
        if isinstance(query, Select):
            return self.execute_select(query, reduce_output=reduce_output)
        elif isinstance(query, Constant):
            return self.execute_constant(query)
        elif isinstance(query, Tuple):
            return pd.Series([self.execute_query(item) for item in query.items])
        else:
            raise QueryExecutionException(f'No idea how to execute query statement {type(query)}')



from dfsql.engine import pd
import re
from dfsql.exceptions import CommandException


class Command:
    name = None
    default_args = {}

    def __init__(self, args):
        self.args = self.substitute_defaults(args)
        self.validate_args(self.args)

    def validate_args(self, args):
        pass

    def substitute_defaults(self, args):
        if args:
            for i, arg in enumerate(args):
                if arg is None and self.default_args.get(i):
                    args[i] = self.default_args[i]
        return args

    @classmethod
    def from_string(cls, text):
        return None

    def execute(self, data_source):
        pass


class CreateTableCommand(Command):
    name = 'CREATE TABLE'
    default_args = {
        1: True, # clean_data
    }

    def validate_args(self, args):
        if len(args) > 2:
            raise CommandException(f"Too many arguments for command {self.name}")

        if not isinstance(args[0], str):
            raise CommandException(f"First argument must be a file path, got instead: {args[0]}.")

        if len(args) > 1 and not isinstance(args[1], bool):
            raise CommandException(f"Second argument (clean_data) must be a boolean, got instead: {args[1]}")

    @classmethod
    def from_string(cls, text):
        if not text.startswith(cls.name):
            return None

        pattern = r'^CREATE TABLE \((\S+)(?:, ((?:True)|(?:False)))?\);?$'

        matches = re.match(pattern, text)
        if not matches:
            return None
        args = [(arg.strip(' \'\"') if arg is not None else None) for arg in matches.groups()]
        args[0] = str(args[0])
        if len(args) > 1 and args[1] is not None:
            args[1] = bool(args[1])
        return cls(args)

    def execute(self, data_source):
        fpath = self.args[0]
        clean = self.args[1]
        data_source.add_table_from_file(fpath, clean=clean)
        return 'OK'


class DropTableCommand(Command):
    name = 'DROP TABLE'

    def validate_args(self, args):
        if not isinstance(args[0], str):
            raise CommandException(f"Expected only argument for {self.name} to be a string table name, got instead: {args[0]}.")

    @classmethod
    def from_string(cls, text):
        if not text.startswith(cls.name):
            return None

        pattern = r'^DROP TABLE (\S+);?$'

        matches = re.match(pattern, text)
        if not matches:
            return None
        args = [(arg.strip(' \'\"') if arg is not None else None) for arg in matches.groups()]
        args[0] = str(args[0])
        return cls(args)

    def execute(self, data_source):
        name = self.args[0]
        data_source.drop_table(name)
        return 'OK'


class ShowTablesCommand(Command):
    name = 'SHOW TABLES'

    def validate_args(self, args):
        if args:
            raise CommandException(f"No arguments expected for command {self.name}")

    @classmethod
    def from_string(cls, text):
        if not text.startswith(cls.name):
            return None

        pattern = r'^SHOW TABLES\s*;?$'

        matches = re.match(pattern, text)
        if not matches:
            return None
        args = None
        return cls(args)

    def execute(self, data_source):
        rows = []
        for tname, table in data_source.tables.items():
            rows.append((table.name, table.fpath))
        return pd.DataFrame(rows, columns=['name', 'fpath'])


command_types = [CreateTableCommand, DropTableCommand, ShowTablesCommand]


def try_parse_command(sql_query):
    for command_type in command_types:
        command = command_type.from_string(sql_query)

        if command:
            return command



import tempfile
import os
import shutil
import time
from dfsql.__about__ import __version__
from dfsql.config import Configuration
from dfsql.exceptions import DfsqlException
from dfsql.data_sources import DataSource
from pandas import DataFrame as PandasDataFrame


def sql_query(sql, *args, ds_kwargs=None, custom_functions=None, **kwargs):
    ds_args = ds_kwargs or {}
    custom_functions = custom_functions or {}
    from_tables = kwargs
    if not from_tables or not isinstance(from_tables, dict):
        raise DfsqlException(f"Wrong from_tables value. Expected to be a dict of table names and dataframes, got: {str(from_tables)}")
    ds = None
    tmpdir = None
    try:
        tmpdir = os.path.join(tempfile.gettempdir(), 'dfsql_temp_' + str(round(time.time())))
        ds = DataSource(*args, metadata_dir=tmpdir, custom_functions=custom_functions, **ds_args)
        for table_name, dataframe in from_tables.items():
            if table_name not in sql:
                raise DfsqlException(f"Table {table_name} found in from_tables, but not in the SQL query.")
            tmp_fpath = os.path.join(tmpdir, f'{table_name}.csv')
            PandasDataFrame(dataframe.values, columns=dataframe.columns, index=dataframe.index).to_csv(tmp_fpath, index=False)
            ds.add_table_from_file(tmp_fpath)

        result = ds.query(sql)
        return result
    finally:
        if ds:
            ds.clear_metadata(ds.metadata_dir)
        if tmpdir:
            shutil.rmtree(tmpdir)

